from logging import Logger, getLogger
from pathlib import Path
from typing import Optional

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      PronunciationDict, SerializationOptions, load_dict, save_dict)

from pronunciation_dictionary_utils_cli.logging_configuration import get_file_logger


def try_save_dict(pronunciation_dict: PronunciationDict, path: Path, encoding: str, options: SerializationOptions, logger: Logger) -> bool:
  try:
    save_dict(pronunciation_dict, path, encoding, options)
  except Exception as ex:
    logger.debug(ex)
    logger.error(f"Dictionary \"{path.absolute()}\" couldn't be written.")
    return False
  return True


def try_load_dict(path: Path, encoding: str, options: DeserializationOptions, mp_options: MultiprocessingOptions, logger: Logger) -> Optional[PronunciationDict]:
  # Output logs to file
  pdict_logger = getLogger("pronunciation_dictionary.deserialization")
  pdict_logger.parent = get_file_logger()

  try:
    result = load_dict(path, encoding, options, mp_options)
  except Exception as ex:
    logger.debug(ex)
    logger.error(f"Dictionary \"{path.absolute()}\" couldn't be read.")
    return None
  return result




from collections import OrderedDict

from pronunciation_dictionary_utils.pronunciations_remove_symbols import \
  remove_symbols_from_pronunciations_entry


def test_component():
  pronunciations = OrderedDict((
    (("ˌaʊ", "ɚ", "s", ".", "ˈe", "l", "v", "z"), 6),
    (("ˌaʊ", "ɚ", "s", ".", "ˈe", "l", "v", "z", "!"), 7),
    ((".", "ˌaʊ", "ɚ", "s", ".", "ˈe", "l", "v", "z", ","), 8),
  ))

  result = remove_symbols_from_pronunciations_entry(pronunciations, {"!", ",", "."}, "both")

  assert result == OrderedDict((
    (("ˌaʊ", "ɚ", "s", ".", "ˈe", "l", "v", "z"), 21),
  ))



from argparse import ArgumentParser, Namespace
from logging import Logger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils import normalize_weights
from pronunciation_dictionary_utils_cli.argparse_helper import (add_io_group, add_mp_group,
                                                                parse_existing_file)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict


def get_weights_normalization_parser(parser: ArgumentParser):
  parser.description = "Normalize weights in dictionary."
  parser.add_argument("dictionary", metavar='DICTIONARY',
                      type=parse_existing_file, help="dictionary file")
  add_io_group(parser)
  add_mp_group(parser)
  return normalize_weights_ns


def normalize_weights_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)

  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)
  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options, logger)
  if dictionary_instance is None:
    return False

  changed_counter = normalize_weights(dictionary_instance, mp_options, silent=False)

  changed_anything = changed_counter > 0

  if not changed_anything:
    logger.info("Didn't change anything.")
    return True

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options, logger)
  if not success:
    return False

  logger.info(f"Written dictionary to: \"{ns.dictionary.absolute()}\".")

  return True



from argparse import ArgumentParser, Namespace
from logging import Logger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils_cli.argparse_helper import (ConvertToOrderedSetAction,
                                                                add_deserialization_group,
                                                                add_mp_group,
                                                                add_serialization_group,
                                                                parse_existing_file)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict


def get_formatting_parser(parser: ArgumentParser):
  parser.description = ""

  parser.add_argument("dictionaries", metavar='DICTIONARY', type=parse_existing_file, nargs="+",
                      help="dictionary files", action=ConvertToOrderedSetAction)
  add_deserialization_group(parser)
  add_serialization_group(parser)
  add_mp_group(parser)

  return adjust_formatting


def adjust_formatting(ns: Namespace, logger: Logger, flogger: Logger):
  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  s_options = SerializationOptions(ns.parts_sep, ns.include_numbers, ns.include_weights)

  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  for dictionary in ns.dictionaries:
    dictionary_instance = try_load_dict(
      dictionary, ns.deserialization_encoding, lp_options, mp_options, logger)
    if dictionary_instance is None:
      return False

    success = try_save_dict(dictionary_instance, dictionary,
                            ns.serialization_encoding, s_options, logger)
    if not success:
      return False

    logger.info(f"Written dictionary to: \"{dictionary.absolute()}\".")
  return True


import argparse
import logging
import platform
import sys
from argparse import ArgumentParser
from importlib.metadata import version
from logging import getLogger
from pathlib import Path
from pkgutil import iter_modules
from tempfile import gettempdir
from time import perf_counter
from typing import Callable, Generator, List, Tuple

from pronunciation_dictionary_utils_cli.argparse_helper import get_optional, parse_path
from pronunciation_dictionary_utils_cli.formatting_adjustment import get_formatting_parser
from pronunciation_dictionary_utils_cli.logging_configuration import (configure_root_logger,
                                                                      get_file_logger,
                                                                      init_and_return_loggers,
                                                                      try_init_file_logger)
from pronunciation_dictionary_utils_cli.merging import get_merging_parser
from pronunciation_dictionary_utils_cli.phoneme_set_extraction import \
  get_phoneme_set_extraction_parser
from pronunciation_dictionary_utils_cli.pronunciations_map_symbols import \
  get_pronunciations_map_symbols_parser
# nat: new functionality
from pronunciation_dictionary_utils_cli.pronunciations_map_symbols_json import \
  get_pronunciations_map_symbols_json_parser
from pronunciation_dictionary_utils_cli.pronunciations_remove_symbols import \
  get_pronunciations_remove_symbols_parser
from pronunciation_dictionary_utils_cli.pronunciations_sorting import \
  get_pronunciations_sorting_parser
from pronunciation_dictionary_utils_cli.single_pronunciation_selection import \
  get_single_pronunciation_selection_parser
from pronunciation_dictionary_utils_cli.subset_extraction import get_subset_extraction_parser
from pronunciation_dictionary_utils_cli.vocabulary_extraction import \
  get_vocabulary_extraction_parser
from pronunciation_dictionary_utils_cli.vocabulary_remove_symbols import \
  get_vocabulary_remove_symbols_parser
from pronunciation_dictionary_utils_cli.weights_normalization import \
  get_weights_normalization_parser
from pronunciation_dictionary_utils_cli.words_casing_adjustment import \
  get_words_casing_adjustment_parser
from pronunciation_dictionary_utils_cli.words_remove_symbols import get_words_remove_symbols_parser
from pronunciation_dictionary_utils_cli.words_sorting import get_words_sorting_parser

prog_name = "pronunciation-dictionary-utils"
__version__ = version(prog_name)

INVOKE_HANDLER_VAR = "invoke_handler"

CONSOLE_PRINT_GREEN = "\x1b[1;49;32m"
CONSOLE_PRINT_RED = "\x1b[1;49;31m"
CONSOLE_PRINT_RESET = "\x1b[0m"


def custom_formatter(prog):
  return argparse.ArgumentDefaultsHelpFormatter(prog, max_help_position=40)


def print_features():
  parsers = get_parsers()
  for command, description, method in parsers:
    print(f"- `{command}`: {description}")


def get_parsers() -> Generator[Tuple[str, str, Callable], None, None]:
  yield from (
    ("export-vocabulary", "export vocabulary from dictionaries",
     get_vocabulary_extraction_parser),
    ("export-phonemes", "export phoneme set from dictionaries",
     get_phoneme_set_extraction_parser),
    ("merge", "merge dictionaries together",
     get_merging_parser),
    ("extract", "extract subset of dictionary vocabulary",
     get_subset_extraction_parser),
    ("map-symbols-in-pronunciations", "map phonemes/symbols in pronunciations to another phoneme/symbol, e.g., mapping ARPAbet to IPA",
     get_pronunciations_map_symbols_parser),
    # nat: new functionality
    ("map-symbols-in-pronunciations-json", "map phonemes/symbols in pronunciations to phoneme/symbol specified in file",
     get_pronunciations_map_symbols_json_parser),
    ("remove-symbols-from-vocabulary", "remove phonemes/symbols from vocabulary",
     get_vocabulary_remove_symbols_parser),
    ("remove-symbols-from-pronunciations", "remove phonemes/symbols from pronunciations",
     get_pronunciations_remove_symbols_parser),
    ("remove-symbols-from-words", "remove characters/symbols from words",
     get_words_remove_symbols_parser),
    ("change-formatting", "change formatting of dictionaries",
     get_formatting_parser),
    ("select-single-pronunciation", "select single pronunciation",
     get_single_pronunciation_selection_parser),
    ("change-word-casing", "transform all words to upper- or lower-case",
     get_words_casing_adjustment_parser),
    ("sort-words", "sort dictionary after words", get_words_sorting_parser),
    ("sort-pronunciations", "sort dictionary pronunciations", get_pronunciations_sorting_parser),
    ("normalize-weights", "normalize pronunciation weights for each word", get_weights_normalization_parser),
  )


def _init_parser():
  main_parser = ArgumentParser(
    formatter_class=custom_formatter,
    description="This program provides methods to modify pronunciation dictionaries.",
  )
  main_parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)
  subparsers = main_parser.add_subparsers(help="description")
  default_log_path = Path(gettempdir()) / f"{prog_name}.log"

  for command, description, method in get_parsers():
    method_parser = subparsers.add_parser(
      command, help=description, formatter_class=custom_formatter)
    # init parser
    invoke_method = method(method_parser)
    method_parser.set_defaults(**{
      INVOKE_HANDLER_VAR: invoke_method,
    })
    logging_group = method_parser.add_argument_group("logging arguments")
    logging_group.add_argument("--log", type=get_optional(parse_path), metavar="FILE",
                               nargs="?", const=None, help="path to write the log", default=default_log_path)
    logging_group.add_argument("--debug", action="store_true",
                               help="include debugging information in log")

  return main_parser


def parse_args(args: List[str]) -> None:
  configure_root_logger()
  root_logger = getLogger()

  local_debugging = debug_file_exists()
  if local_debugging:
    root_logger.debug(f"Received arguments: {str(args)}")

  parser = _init_parser()

  try:
    ns = parser.parse_args(args)
  except SystemExit as error:
    error_code = error.args[0]
    # -v -> 0; invalid arg -> 2
    sys.exit(error_code)

  if local_debugging:
    root_logger.debug(f"Parsed arguments: {str(ns)}")

  if not hasattr(ns, INVOKE_HANDLER_VAR):
    parser.print_help()
    sys.exit(0)

  invoke_handler: Callable[..., bool] = getattr(ns, INVOKE_HANDLER_VAR)
  delattr(ns, INVOKE_HANDLER_VAR)
  log_to_file = ns.log is not None
  if log_to_file:
    log_level = logging.DEBUG if (local_debugging or ns.debug) else logging.INFO
    log_to_file_successful = try_init_file_logger(ns.log, log_level)
    if not log_to_file_successful:
      root_logger.warning("Logging to file is not possible!")

  flogger = get_file_logger()
  if not local_debugging:
    sys_version = sys.version.replace('\n', '')
    flogger.debug(f"CLI version: {__version__}")
    flogger.debug(f"Python version: {sys_version}")
    flogger.debug("Modules: %s", ', '.join(sorted(p.name for p in iter_modules())))

    my_system = platform.uname()
    flogger.debug(f"System: {my_system.system}")
    flogger.debug(f"Node Name: {my_system.node}")
    flogger.debug(f"Release: {my_system.release}")
    flogger.debug(f"Version: {my_system.version}")
    flogger.debug(f"Machine: {my_system.machine}")
    flogger.debug(f"Processor: {my_system.processor}")

  flogger.debug(f"Received arguments: {str(args)}")
  flogger.debug(f"Parsed arguments: {str(ns)}")

  start_time = perf_counter()
  cmd_flogger, cmd_logger = init_and_return_loggers(__name__)
  #success, changed_anything = invoke_handler(ns, cmd_logger, cmd_flogger)
  success = invoke_handler(ns, cmd_logger, cmd_flogger)
  if success:
    root_logger.info(f"{CONSOLE_PRINT_GREEN}Everything was successful!{CONSOLE_PRINT_RESET}")
    flogger.info("Everything was successful!")
  else:
    if log_to_file:
      root_logger.error("Not everything was successful! See log for details.")
    else:
      root_logger.error("Not everything was successful!")
    flogger.error("Not everything was successful!")

  duration_s = perf_counter() - start_time
  flogger.debug(f"Total duration (s): {duration_s}")

  if log_to_file:
    # path not encapsulated in "" because it is only console out
    root_logger.info(f"Written log to: {ns.log.absolute()}")

  if not success:
    sys.exit(1)
  sys.exit(0)

def run():
  arguments = sys.argv[1:]
  parse_args(arguments)


def debug_file_exists():
  return (Path(gettempdir()) / f"{prog_name}-debug").is_file()


def create_debug_file():
  if not debug_file_exists():
    (Path(gettempdir()) / f"{prog_name}-debug").write_text("", "UTF-8")


if __name__ == "__main__":
  # create_debug_file()
  # print_features()
  run()



from collections import OrderedDict

from pronunciation_dictionary_utils.words_sorting import sort_words


def test_component_ascending_cased():
  pronunciation_dict = OrderedDict()

  pronunciation_dict["B"] = OrderedDict((
    (("b",), 1),
  ))

  pronunciation_dict["b"] = OrderedDict((
    (("b",), 1),
  ))

  pronunciation_dict["A"] = OrderedDict((
    (("a",), 1),
  ))

  pronunciation_dict["a"] = OrderedDict((
    (("a",), 1),
  ))

  pronunciation_dict["C"] = OrderedDict((
    (("c",), 1),
  ))

  pronunciation_dict["c"] = OrderedDict((
    (("c",), 1),
  ))

  result = sort_words(pronunciation_dict, False, True)

  assert list(result.keys()) == ["A", "B", "C", "a", "b", "c"]
  assert result["A"] == OrderedDict((
    (("a",), 1),
  ))


def test_component_ascending_uncased():
  pronunciation_dict = OrderedDict()

  pronunciation_dict["B"] = OrderedDict((
    (("b",), 1),
  ))

  pronunciation_dict["b"] = OrderedDict((
    (("b",), 1),
  ))

  pronunciation_dict["A"] = OrderedDict((
    (("a",), 1),
  ))

  pronunciation_dict["a"] = OrderedDict((
    (("a",), 1),
  ))

  pronunciation_dict["C"] = OrderedDict((
    (("c",), 1),
  ))

  pronunciation_dict["c"] = OrderedDict((
    (("c",), 1),
  ))

  result = sort_words(pronunciation_dict, False, False)

  assert list(result.keys()) == ["A", "a", "B", "b", "C", "c"]
  assert result["A"] == OrderedDict((
    (("a",), 1),
  ))


def test_component_descending_cased():
  pronunciation_dict = OrderedDict()

  pronunciation_dict["B"] = OrderedDict((
    (("b",), 1),
  ))

  pronunciation_dict["b"] = OrderedDict((
    (("b",), 1),
  ))

  pronunciation_dict["A"] = OrderedDict((
    (("a",), 1),
  ))

  pronunciation_dict["a"] = OrderedDict((
    (("a",), 1),
  ))

  pronunciation_dict["C"] = OrderedDict((
    (("c",), 1),
  ))

  pronunciation_dict["c"] = OrderedDict((
    (("c",), 1),
  ))

  result = sort_words(pronunciation_dict, True, True)

  assert list(result.keys()) == ["c", "b", "a", "C", "B", "A"]
  assert result["A"] == OrderedDict((
    (("a",), 1),
  ))


def test_component_descending_uncased():
  pronunciation_dict = OrderedDict()

  pronunciation_dict["B"] = OrderedDict((
    (("b",), 1),
  ))

  pronunciation_dict["b"] = OrderedDict((
    (("b",), 1),
  ))

  pronunciation_dict["A"] = OrderedDict((
    (("a",), 1),
  ))

  pronunciation_dict["a"] = OrderedDict((
    (("a",), 1),
  ))

  pronunciation_dict["C"] = OrderedDict((
    (("c",), 1),
  ))

  pronunciation_dict["c"] = OrderedDict((
    (("c",), 1),
  ))

  result = sort_words(pronunciation_dict, True, False)

  assert list(result.keys()) == ["c", "C", "b", "B", "a", "A"]
  assert result["A"] == OrderedDict((
    (("a",), 1),
  ))



from collections import OrderedDict
from functools import partial
from multiprocessing.pool import Pool
from typing import Literal, Optional, Tuple

from ordered_set import OrderedSet
from pronunciation_dictionary import MultiprocessingOptions, PronunciationDict, Word
from tqdm import tqdm

from pronunciation_dictionary_utils.common import merge_pronunciations
from pronunciation_dictionary_utils.validation import validate_dictionary


def __validate_mode(mode: str) -> Optional[str]:
  if mode not in ["lower", "upper"]:
    return "Invalid value!"
  return None


def change_word_casing(dictionary: PronunciationDict, mode: str, mp_options: MultiprocessingOptions, silent: bool = False) -> int:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := __validate_mode(mode):
    raise ValueError(f"Parameter 'mode': {msg}")

  process_method = partial(
    __process_change_casing,
    mode=mode,
  )

  with Pool(
    processes=mp_options.n_jobs,
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    iterator = pool.imap(process_method, dictionary.keys(), mp_options.chunksize)
    new_words_to_words = dict(tqdm(iterator, total=len(dictionary), unit="words", disable=silent))

  new_words = OrderedDict((
    (k, new_word)
    for k in dictionary.keys()
    if (new_word := new_words_to_words[k])
  ))

  del new_words_to_words

  created_words = OrderedSet()

  for word, new_word in new_words.items():
    popped_pronunciations = dictionary.pop(word)
    if new_word in dictionary:
      existing_pronunciations = dictionary[new_word]
      merge_pronunciations(existing_pronunciations, popped_pronunciations)
    else:
      created_words.add(new_word)
      dictionary[new_word] = popped_pronunciations

  removed_words = OrderedSet(new_words.keys())
  return removed_words, created_words


def __process_change_casing(word: Word, mode: Literal["upper", "lower"]) -> Tuple[Word, Optional[Word]]:
  if mode == "upper":
    new_word = word.upper()
  elif mode == "lower":
    new_word = word.lower()
  else:
    assert False

  if new_word != word:
    return word, new_word
  return word, None



from pronunciation_dictionary_utils.pronunciations_remove_symbols import remove_symbols_all


def test_component():
  symbols = (".", ";", "a", "d", ".", "b", "c", ".", "!")
  result = tuple(remove_symbols_all(symbols, {".", ";", "!"}))
  assert result == ("a", "d", "b", "c")



from argparse import ArgumentParser, Namespace
from logging import Logger
from pathlib import Path
from tempfile import gettempdir

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils import remove_symbols_from_pronunciations
from pronunciation_dictionary_utils_cli.argparse_helper import (ConvertToOrderedSetAction,
                                                                add_io_group, add_mp_group,
                                                                get_optional, parse_existing_file,
                                                                parse_non_empty_or_whitespace,
                                                                parse_path)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict

DEFAULT_EMPTY_WEIGHT = 1


def get_pronunciations_remove_symbols_parser(parser: ArgumentParser):
  default_removed_out = Path(gettempdir()) / "removed-words.txt"
  parser.description = "Remove symbols from pronunciations."
  parser.add_argument("dictionary", metavar='DICTIONARY',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("symbols", type=str, metavar='SYMBOL', nargs='+',
                      help="remove these symbols from the pronunciations", action=ConvertToOrderedSetAction)
  parser.add_argument("-m", "--mode", type=str, choices=["all", "start", "end", "both"], metavar="MODE",
                      help="mode to remove the symbols: all = on all locations; start = only from start; end = only from end; both = start + end", default="both")
  parser.add_argument("-k", "--keep-empty", action="store_true",
                      help="if a pronunciation will be empty after removal, keep the corresponding word in the dictionary and assign the value of empty-symbol")
  parser.add_argument("-es", "--empty-symbol", metavar="SYMBOL", type=get_optional(parse_non_empty_or_whitespace),
                      help="if keep-empty: assign this symbol to the word where no pronunciations result because of the symbol removal", default="sil")
  parser.add_argument("-ro", "--removed-out", metavar="PATH", type=get_optional(parse_path),
                      help="write removed words (i.e., words that had no pronunciation anymore) to this file", default=default_removed_out)
  add_io_group(parser)
  add_mp_group(parser)
  return remove_symbols_from_pronunciations_ns


def remove_symbols_from_pronunciations_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  if ns.keep_empty and ns.empty_symbol is None:
    logger.error("An empty symbol needs to be supplied if keep-empty is true!")
    return False

  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options, logger)
  if dictionary_instance is None:
    return False

  removed_words, changed_counter = remove_symbols_from_pronunciations(
    dictionary_instance, ns.symbols, ns.mode, ns.keep_empty, ns.empty_symbol, mp_options, silent=False)

  if changed_counter == 0:
    logger.info("Didn't change anything.")
    return True

  logger.info(f"Changed pronunciations of {changed_counter} word(s).")

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options, logger)
  if not success:
    return False

  logger.info(f"Written dictionary to: \"{ns.dictionary.absolute()}\"")

  if len(removed_words) > 0:
    logger.warning(f"{len(removed_words)} words were removed.")
    if ns.removed_out is not None:
      content = "\n".join(removed_words)
      ns.removed_out.parent.mkdir(parents=True, exist_ok=True)
      try:
        ns.removed_out.write_text(content, "UTF-8")
      except Exception as ex:
        logger.debug(ex)
        logger.error("Removed words output couldn't be created!")
        return False
      logger.info(f"Written removed words to: \"{ns.removed_out.absolute()}\".")
  else:
    logger.info("No words were removed.")
  return True



from pronunciation_dictionary_utils.pronunciations_map_symbols_dict import get_mappable_symbols


def test_symbols_found():
    sounds_in_dictionary = {"AO", "AO2", "AO3", "AA1", "EY2", ".", "A03", "NN", "HH"}
    sounds_in_mappings = {"EY1", "EY", "AO1", "EY0", "AO", "AO0", "AO2", "EY2"}

    expected_mappable_symbols = {"AO2", "EY2", "AO"}

    mappable_symbols = get_mappable_symbols(sounds_in_dictionary, sounds_in_mappings)

    assert mappable_symbols == expected_mappable_symbols


def test_symbols_not_found():
    sounds_in_dictionary = {"AO", "AO2", "AO3", "AA1", "EY2", ".", "A03", "NN", "HH"}
    sounds_in_mappings = {"EY1", "EY", "AO1", "EY0", "AO0"}

    mappable_symbols = get_mappable_symbols(sounds_in_dictionary, sounds_in_mappings)

    assert mappable_symbols == set()


def test_empty():
    sounds_in_dictionary = set()
    sounds_in_mappings = set()

    mappable_symbols = get_mappable_symbols(sounds_in_dictionary, sounds_in_mappings)

    assert mappable_symbols == set()


from pronunciation_dictionary_utils.common import merge_pronunciations
from pronunciation_dictionary_utils.merging import merge_dictionaries
from pronunciation_dictionary_utils.pronunciations_map_symbols import map_symbols
from pronunciation_dictionary_utils.pronunciations_map_symbols_dict import map_symbols_dict
from pronunciation_dictionary_utils.pronunciations_remove_symbols import \
  remove_symbols_from_pronunciations
from pronunciation_dictionary_utils.pronunciations_replace_pronunciation import \
  replace_symbols_in_pronunciations
from pronunciation_dictionary_utils.pronunciations_sorting import sort_pronunciations
from pronunciation_dictionary_utils.single_pronunciation_selection import \
  select_single_pronunciation
from pronunciation_dictionary_utils.subset_extraction import select_subset_dictionary
from pronunciation_dictionary_utils.weights_normalization import normalize_weights
from pronunciation_dictionary_utils.words_casing_adjustment import change_word_casing
from pronunciation_dictionary_utils.words_remove_symbols import remove_symbols_from_words
from pronunciation_dictionary_utils.words_sorting import sort_words



from argparse import ArgumentParser, Namespace
from logging import Logger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils import select_single_pronunciation
from pronunciation_dictionary_utils_cli.argparse_helper import (add_io_group, add_mp_group,
                                                                get_optional, parse_existing_file,
                                                                parse_non_negative_integer)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict


def get_single_pronunciation_selection_parser(parser: ArgumentParser):
  parser.description = "Select a single pronunciation for each word."
  parser.add_argument("dictionary", metavar='DICTIONARY',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("-m", "--mode", type=str, choices=["first", "last", "highest-weight", "lowest-weight", "shortest", "longest", "random", "weighted"], metavar="MODE",
                      help="mode to select the target pronunciation: first -> the first pronunciation; last -> the last pronunciation; highest-weight -> the first pronunciation with the highest weight; lowest-weight -> the first pronunciation with the lowest-weight; shortest -> the pronunciation with the least amount of phonemes; longest -> the pronunciation with the most amount of phonemes; random -> a random pronunciation & weighted -> same as random but higher weights result in higher probabilities", default="first")
  parser.add_argument("--seed", type=get_optional(parse_non_negative_integer),
                      metavar="SEED", help="custom seed if mode is random or weight", default=None)
  add_io_group(parser)
  add_mp_group(parser)
  return remove_multiple_pronunciations_ns


def remove_multiple_pronunciations_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  lp_options = DeserializationOptions(
    ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options, logger)
  if dictionary_instance is None:
    return False

  changed_counter = select_single_pronunciation(
    dictionary_instance, ns.mode, ns.seed, mp_options, silent=False)

  if changed_counter == 0:
    logger.info("Didn't change anything.")
    return True

  logger.info(f"Changed pronunciations of {changed_counter} word(s).")

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options, logger)
  if not success:
    return False

  logger.info(f"Written dictionary to: \"{ns.dictionary.absolute()}\".")
  return True



from collections import OrderedDict

from pronunciation_dictionary import MultiprocessingOptions

from pronunciation_dictionary_utils.pronunciations_map_symbols_dict import apply_mapping_full


def test_with_changes() -> None:
    test_dictionary = OrderedDict([("test", OrderedDict([(("AO2",), 1)]))])
    mappings = {"AO2": "ˌɔ"}
    mappable_symbol = "AO2"
    expected_result = OrderedDict([("test", OrderedDict([(("ˌɔ",), 1)]))])
    mp_options = MultiprocessingOptions(n_jobs=4, maxtasksperchild=100, chunksize=10)

    changed_words = apply_mapping_full(test_dictionary, mappings, mappable_symbol, mp_options)

    assert changed_words == {"test"}
    assert test_dictionary == expected_result


def test_with_whitespaces() -> None:
    test_dictionary = OrderedDict([("test", OrderedDict([(("EY2",), 1)]))])
    mappings = {"EY2": "ˌe ɪ"}
    mappable_symbol = ("EY2")
    expected_result = OrderedDict([("test", OrderedDict([(("ˌe", "ɪ"), 1)]))])
    mp_options = MultiprocessingOptions(n_jobs=4, maxtasksperchild=100, chunksize=10)

    changed_words = apply_mapping_full(test_dictionary, mappings, mappable_symbol, mp_options)

    assert changed_words == {"test"}
    assert test_dictionary == expected_result



from collections import OrderedDict
from typing import Any, Optional

from ordered_set import OrderedSet
from pronunciation_dictionary import MultiprocessingOptions, PronunciationDict, Pronunciations, Word


class ValidationError():
  # pylint: disable=no-self-use
  @property
  def default_message(self) -> str:
    return ""


class InternalError(ValidationError):
  @property
  def default_message(self) -> str:
    return "Internal error!"


def validate_vocabulary(vocabulary: OrderedSet[Word]) -> Optional[str]:
  if not isinstance(vocabulary, OrderedSet):
    return "Type needs to be 'OrderedSet'!"
  return None


def validate_dictionary(dictionary: PronunciationDict) -> Optional[str]:
  return validate_type(dictionary, OrderedDict)


def validate_mp_options(mp_options: MultiprocessingOptions) -> Optional[str]:
  if not (isinstance(mp_options.chunksize, int) and mp_options.chunksize > 0):
    return "Property 'chunksize' is invalid!"
  if not (mp_options.maxtasksperchild is None or (isinstance(mp_options.maxtasksperchild, int) and mp_options.maxtasksperchild > 0)):
    return "Property 'maxtasksperchild' is invalid!"
  if not (isinstance(mp_options.n_jobs, int) and mp_options.n_jobs > 0):
    return "Property 'n_jobs' is invalid!"
  return None


def validate_seed(seed: int) -> Optional[str]:
  if not 0 < seed and not isinstance(seed, int):
    return "Invalid value!"
  return None


def validate_type(obj: Any, t: type) -> Optional[str]:
  if not isinstance(obj, t):
    return f"Value needs of type '{t.__name__}'!"
  return None


def validate_pronunciations(pronunciations: Pronunciations) -> Optional[str]:
  if msg := validate_type(pronunciations, OrderedDict):
    return msg
  if not len(pronunciations) > 0:
    return "At least one pronunciation is required!"
  return None



from collections import OrderedDict

from pronunciation_dictionary import PronunciationDict

from pronunciation_dictionary_utils.validation import validate_dictionary, validate_type


def sort_words(dictionary: PronunciationDict, descending: bool, consider_case: bool) -> PronunciationDict:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_type(descending, bool):
    raise ValueError(f"Parameter 'descending': {msg}")
  if msg := validate_type(consider_case, bool):
    raise ValueError(f"Parameter 'consider_case': {msg}")

  words = dictionary.keys()
  sorted_words = sorted(words, reverse=descending,
                        key=lambda w: w if consider_case else (w.lower(), w))
  result = OrderedDict((
    (word, dictionary[word]) for word in sorted_words
  ))
  return result



from pronunciation_dictionary_utils.pronunciations_map_symbols_dict import get_unmappable_symbols


def test_symbols_found():
  sounds_in_dictionary = {"AO", "AO2", "AO3", "AA1", "EY2", ".", "A03", "NN", "HH"}
  sounds_in_mappings = {"EY1", "EY", "AO1", "EY0", "AO", "AO0", "AO2", "EY2"}

  expected_unmappable_symbols = {"A03", "AA1", "AO3", "HH", "NN", "."}
  unmappable_symbols = get_unmappable_symbols(sounds_in_dictionary, sounds_in_mappings)

  assert unmappable_symbols == expected_unmappable_symbols


def test_symbols_not_found():
  sounds_in_dictionary = {"AO", "AO2", "EY2"}
  sounds_in_mappings = {"EY1", "EY", "AO1", "EY0", "AO", "AO0", "AO2", "EY2"}

  unmappable_symbols = get_unmappable_symbols(sounds_in_dictionary, sounds_in_mappings)

  assert unmappable_symbols == set()


def test_empty():
  sounds_in_dictionary = set()
  sounds_in_mappings = set()

  unmappable_symbols = get_unmappable_symbols(sounds_in_dictionary, sounds_in_mappings)

  assert unmappable_symbols == set()



from argparse import ArgumentParser, Namespace
from logging import Logger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils import merge_dictionaries
from pronunciation_dictionary_utils_cli.argparse_helper import (ConvertToOrderedSetAction,
                                                                add_io_group, add_mp_group,
                                                                parse_existing_file)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict


def get_merging_parser(parser: ArgumentParser):
  parser.description = "Merge multiple dictionaries into one."
  parser.add_argument("dictionary", metavar='DICTIONARY',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("dictionaries", metavar='MERGE-DICTIONARY', type=parse_existing_file, nargs="+",
                      help="dictionary files that should be merged to DICTIONARY", action=ConvertToOrderedSetAction)
  parser.add_argument("--duplicate-handling", type=str, metavar="MODE",
                      choices=["add", "extend", "replace"], help="sets how existing pronunciations should be handled: add = add missing pronunciations; extend = add missing pronunciations and extend existing ones; replace: add missing pronunciations and replace existing ones.", default="extend")
  add_io_group(parser)
  add_mp_group(parser)
  return merge_dictionary_files_ns


def merge_dictionary_files_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  assert len(ns.dictionaries) > 0

  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  resulting_dictionary = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options, logger)
  if resulting_dictionary is None:
    return False

  changed_anything = False
  for dictionary in ns.dictionaries:
    dictionary_instance = try_load_dict(dictionary, ns.encoding, lp_options, mp_options, logger)
    if dictionary_instance is None:
      return False
    changed_anything |= merge_dictionaries(
      resulting_dictionary, dictionary_instance, ns.duplicate_handling)

  if not changed_anything:
    logger.info("Didn't change anything.")
    return True

  success = try_save_dict(resulting_dictionary, ns.dictionary,
                          ns.encoding, s_options, logger)
  if not success:
    return False

  logger.info(f"Written dictionary to: \"{ns.dictionary.absolute()}\".")
  return True



from collections import OrderedDict

import pytest

from pronunciation_dictionary_utils.common import merge_pronunciations


def test_component():
  p1 = OrderedDict()
  p1[("a",)] = 2
  p1[("b",)] = 3
  p1[("d",)] = 5

  p2 = OrderedDict()
  p2[("a",)] = 2
  p2[("c",)] = 4
  p2[("e", "f")] = 6

  changed_anything = merge_pronunciations(p1, p2)

  assert changed_anything
  assert p1 == OrderedDict((
    (("a",), 4),
    (("b",), 3),
    (("d",), 5),
    (("c",), 4),
    (("e", "f"), 6),
  ))


def test_same_instance__changes_nothing():
  p1 = OrderedDict()
  p1[("a",)] = 2

  changed_anything = merge_pronunciations(p1, p1)

  assert not changed_anything
  assert p1 == OrderedDict((
    (("a",), 2),
  ))


def test_same_key__weights_are_added():
  p1 = OrderedDict()
  p1[("a",)] = 2

  p2 = OrderedDict()
  p2[("a",)] = 3

  changed_anything = merge_pronunciations(p1, p2)

  assert changed_anything
  assert p1 == OrderedDict((
    (("a",), 5),
  ))


def test_same_key_zero_weight__changes_nothing():
  p1 = OrderedDict()
  p1[("a",)] = 2

  p2 = OrderedDict()
  p2[("a",)] = 0

  changed_anything = merge_pronunciations(p1, p2)

  assert not changed_anything
  assert p1 == OrderedDict((
    (("a",), 2),
  ))


def test_different_key__key_is_added():
  p1 = OrderedDict()
  p1[("a",)] = 2

  p2 = OrderedDict()
  p2[("b",)] = 3

  changed_anything = merge_pronunciations(p1, p2)

  assert changed_anything
  assert p1 == OrderedDict((
    (("a",), 2),
    (("b",), 3),
  ))


def test_p1_empty_raises_error():
  p1 = OrderedDict()

  p2 = OrderedDict()
  p2[("a",)] = 2

  with pytest.raises(ValueError) as error:
    merge_pronunciations(p1, p2)


def test_p2_empty_raises_error():
  p1 = OrderedDict()
  p1[("a",)] = 2

  p2 = OrderedDict()

  with pytest.raises(ValueError) as error:
    merge_pronunciations(p1, p2)



from typing import Optional

from ordered_set import OrderedSet
from pronunciation_dictionary import PronunciationDict

from pronunciation_dictionary_utils.common import merge_pronunciations
from pronunciation_dictionary_utils.validation import validate_dictionary


def __validate_mode(mode: str) -> Optional[str]:
  if mode not in ["add", "replace", "extend"]:
    return "Invalid value!"
  return None


def merge_dictionaries(dictionary: PronunciationDict, other_dictionary: PronunciationDict, mode: str) -> bool:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_dictionary(other_dictionary):
    raise ValueError(f"Parameter 'other_dictionary': {msg}")
  if msg := __validate_mode(mode):
    raise ValueError(f"Parameter 'mode': {msg}")

  changed_anything = False
  if mode == "add":
    changed_anything = dictionary_add_new(dictionary, other_dictionary)
  elif mode == "replace":
    changed_anything = dictionary_replace(dictionary, other_dictionary)
  elif mode == "extend":
    changed_anything = dictionary_extend(dictionary, other_dictionary)
  else:
    assert False
  return changed_anything


def dictionary_replace(dictionary1: PronunciationDict, dictionary2: PronunciationDict) -> bool:
  changed_anything = False
  for key, value in dictionary2.items():
    if key not in dictionary1:
      changed_anything = True
      dictionary1[key] = value
    else:
      if dictionary1[key] != value:
        dictionary1[key] = value
        changed_anything = True
  return changed_anything


def dictionary_add_new(dictionary1: PronunciationDict, dictionary2: PronunciationDict) -> bool:
  new_keys = OrderedSet(dictionary2.keys()).difference(dictionary1.keys())
  for key in new_keys:
    assert key not in dictionary1
    dictionary1[key] = dictionary2[key]
  changed_anything = len(new_keys) > 0
  return changed_anything


def dictionary_extend(dictionary1: PronunciationDict, dictionary2: PronunciationDict) -> bool:
  keys = OrderedSet(dictionary2.keys())
  same_keys = keys.intersection(dictionary1.keys())
  new_keys = keys.difference(dictionary1.keys())

  changed_anything = False
  for key in same_keys:
    assert key in dictionary1
    changed_anything |= merge_pronunciations(dictionary1[key], dictionary2[key])

  for key in new_keys:
    assert key not in dictionary1
    dictionary1[key] = dictionary2[key]
  changed_anything |= len(new_keys) > 0
  return changed_anything



from collections import OrderedDict

import pytest
from pronunciation_dictionary import MultiprocessingOptions

from pronunciation_dictionary_utils.pronunciations_map_symbols_dict import apply_mapping_partial


def test_with_changes() -> None:
    test_dictionary = OrderedDict([("test", OrderedDict([(("AO2",), 1)]))])
    mappings = {"AO": "ɔ"}
    expected_result = OrderedDict([("test", OrderedDict([(("ɔ2",), 1)]))])
    mappable_symbol = "AO"
    mp_options = MultiprocessingOptions(n_jobs=4, maxtasksperchild=100, chunksize=10)

    changed_words = apply_mapping_partial(test_dictionary, mappings, mappable_symbol, mp_options)

    assert changed_words == {"test"}
    assert test_dictionary == expected_result


def test_with_whitespaces() -> None:
    test_dictionary = OrderedDict([("test", OrderedDict([(("EY2",), 1)]))])
    mappings = {"EY2": "ˌe ɪ"}
    mappable_symbol = "EY2"
    mp_options = MultiprocessingOptions(n_jobs=4, maxtasksperchild=100, chunksize=10)

    with pytest.raises(Exception) as expected_exception:
        apply_mapping_partial(test_dictionary, mappings, mappable_symbol, mp_options)
    assert str(expected_exception.value) == "Whitespaces in mappings aren't supported with partial mapping."



from collections import OrderedDict
from functools import partial
from multiprocessing.pool import Pool
from typing import List, Optional, Set, Tuple, Union

from ordered_set import OrderedSet
from pronunciation_dictionary import (MultiprocessingOptions, PronunciationDict, Pronunciations,
                                      Symbol, Word)
from tqdm import tqdm

from pronunciation_dictionary_utils.validation import (validate_dictionary, validate_mp_options,
                                                       validate_type)


def __validate_symbol(symbol: str) -> Optional[str]:
  if not isinstance(symbol, str):
    return "Value needs of type 'str'!"
  assert len(symbol) > 0
  return None


def map_symbols(dictionary: PronunciationDict, symbols: OrderedSet[Symbol], map_to: Union[List[Symbol], Symbol], partial_mapping: bool, mp_options: MultiprocessingOptions, silent: bool = False) -> Set[str]:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_type(symbols, OrderedSet):
    raise ValueError(f"Parameter 'symbols': {msg}")
  for symbol in symbols:
    if msg := __validate_symbol(symbol):
      raise ValueError(f"Parameter 'symbols': {msg}")
  if len(symbols) == 0:
    raise ValueError("Parameter 'symbols': At least one symbol needs to be passed!")
  if msg := validate_type(partial_mapping, bool):
    raise ValueError(f"Parameter 'partial_mapping': {msg}")
  if partial_mapping:
    if msg := __validate_symbol(map_to):
      raise ValueError(f"Parameter 'map_to': {msg}")
  else:
    if msg := validate_type(map_to, list):
      raise ValueError(f"Parameter 'map_to': {msg}")
    for symbol in map_to:
      if msg := __validate_symbol(symbol):
        raise ValueError(f"Parameter 'map_to': {msg}")
  if msg := validate_mp_options(mp_options):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  if partial_mapping:
    assert isinstance(map_to, str)
    process_method = partial(
      process_map_pronunciations_partial,
      symbols=symbols,
      map_symbol=map_to,
    )
  else:
    assert isinstance(map_to, list)
    process_method = partial(
      process_map_pronunciations_full,
      symbols=symbols,
      mapping_symbols=map_to,
    )

  with Pool(
    processes=mp_options.n_jobs,
    initializer=__init_pool,
    initargs=(dictionary,),
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    all_words = OrderedSet(dictionary.keys())
    iterator = tqdm(pool.imap(process_method, all_words, mp_options.chunksize),
                    total=len(all_words), unit="words", disable=silent)
    new_pronunciations_to_words = dict(iterator)

  changed_words = set()

  for word, new_pronunciations in new_pronunciations_to_words.items():
    changed_pronunciations = new_pronunciations is not None
    if changed_pronunciations:
      dictionary[word] = new_pronunciations
      changed_words.add(word)

  return changed_words


PROCESS_LOOKUP_DICT: PronunciationDict = None


def __init_pool(lookup_dict: PronunciationDict) -> None:
  global PROCESS_LOOKUP_DICT
  PROCESS_LOOKUP_DICT = lookup_dict


def replace_str(s: str, replace: OrderedSet[str], replace_with: str) -> str:
  for r_str in replace:
    if r_str in s:
      s = s.replace(r_str, replace_with)
  return s


def process_map_pronunciations_partial(word: Word, symbols: OrderedSet[Symbol], map_symbol: Symbol) -> Tuple[Word, Optional[Pronunciations]]:
  global PROCESS_LOOKUP_DICT
  assert word in PROCESS_LOOKUP_DICT
  pronunciations = PROCESS_LOOKUP_DICT[word]
  new_pronunciations = map_pronunciations_partial(pronunciations, symbols, map_symbol)
  if new_pronunciations == pronunciations:
    del pronunciations
    del new_pronunciations
    return word, None
  del pronunciations
  return word, new_pronunciations


def process_map_pronunciations_full(word: Word, symbols: OrderedSet[Symbol], mapping_symbols: List[Symbol]) -> Tuple[Word, Optional[Pronunciations]]:
  global PROCESS_LOOKUP_DICT
  assert word in PROCESS_LOOKUP_DICT
  pronunciations = PROCESS_LOOKUP_DICT[word]
  new_pronunciations = map_pronunciations_full(pronunciations, symbols, mapping_symbols)
  if new_pronunciations == pronunciations:
    del pronunciations
    del new_pronunciations
    return word, None
  del pronunciations
  return word, new_pronunciations


def map_pronunciations_partial(pronunciations: Pronunciations, replace_symbols: OrderedSet[Symbol], map_symbol: Symbol) -> Pronunciations:
  assert len(pronunciations) > 0
  assert map_symbol != ""
  new_pronunciations = OrderedDict()
  for pronunciation, weight in pronunciations.items():
    assert len(pronunciation) > 0
    new_pronunciation = tuple(
      replace_str(symbol, replace_symbols, map_symbol)
      for symbol in pronunciation
    )
    if new_pronunciation in new_pronunciations:
      new_pronunciations[new_pronunciation] += weight
    else:
      new_pronunciations[new_pronunciation] = weight
  return new_pronunciations


def map_pronunciations_full(pronunciations: Pronunciations, replace_symbols: OrderedSet[Symbol], mapping_symbols: List[Symbol]) -> Pronunciations:
  assert len(pronunciations) > 0
  assert mapping_symbols != ""
  new_pronunciations = OrderedDict()
  for pronunciation, weight in pronunciations.items():
    assert len(pronunciation) > 0
    new_pronunciation = []
    for symbol in pronunciation:
      if symbol in replace_symbols:
        new_pronunciation.extend(mapping_symbols)
      else:
        new_pronunciation.append(symbol)
    new_pronunciation = tuple(new_pronunciation)
    if new_pronunciation in new_pronunciations:
      new_pronunciations[new_pronunciation] += weight
    else:
      new_pronunciations[new_pronunciation] = weight
  return new_pronunciations



from argparse import ArgumentParser, Namespace
from logging import Logger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils import map_symbols
from pronunciation_dictionary_utils_cli.argparse_helper import (ConvertToOrderedSetAction,
                                                                add_io_group, add_mp_group,
                                                                parse_existing_file,
                                                                parse_non_empty_or_whitespace)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict

DEFAULT_EMPTY_WEIGHT = 1.0


def get_pronunciations_map_symbols_parser(parser: ArgumentParser):
  parser.description = "Map symbols in pronunciations."
  parser.add_argument("dictionary", metavar='DICTIONARY',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("from_symbols", type=parse_non_empty_or_whitespace, metavar='FROM-SYMBOL', nargs='+',
                      help="map these symbols from the pronunciations to MAP_SYMBOL", action=ConvertToOrderedSetAction)
  parser.add_argument("to_symbol", type=parse_non_empty_or_whitespace,
                      metavar="TO-SYMBOL", help="this symbol will be taken")
  parser.add_argument("-pm", "--partial-mapping", action="store_true",
                      help="map symbols inside a symbol; useful when mapping the same vowel having different tones or stress within one operation")
  add_io_group(parser)
  add_mp_group(parser)
  return map_symbols_in_pronunciations_ns


def map_symbols_in_pronunciations_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  lp_options = DeserializationOptions(
    ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options, logger)
  if dictionary_instance is None:
    return False

  to_symbol = ns.to_symbol if ns.partial_mapping else [ns.to_symbol]
  changed_words = map_symbols(
    dictionary_instance, ns.from_symbols, to_symbol, ns.partial_mapping, mp_options, silent=False)

  if len(changed_words) == 0:
    logger.info("Didn't change anything.")
    return True

  logger.info(f"Changed pronunciations of {len(changed_words)} word(s).")

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options, logger)
  if not success:
    return False

  logger.info(f"Written dictionary to: \"{ns.dictionary.absolute()}\"")

  return True



from argparse import ArgumentParser, Namespace
from logging import Logger
from pathlib import Path
from typing import cast

from ordered_set import OrderedSet
from pronunciation_dictionary import DeserializationOptions, MultiprocessingOptions

from pronunciation_dictionary_utils_cli.argparse_helper import (ConvertToOrderedSetAction,
                                                                add_deserialization_group,
                                                                add_encoding_argument, add_mp_group,
                                                                parse_existing_file, parse_path)
from pronunciation_dictionary_utils_cli.io import try_load_dict


def get_vocabulary_extraction_parser(parser: ArgumentParser):
  parser.description = "Export vocabulary (i.e., words that are contained in the dictionaries)."
  parser.add_argument("dictionaries", metavar='DICTIONARY', type=parse_existing_file, nargs="+",
                      help="dictionary files", action=ConvertToOrderedSetAction)
  parser.add_argument("output", metavar="OUTPUT-VOCABULARY", type=parse_path,
                      help="output vocabulary to this file")
  add_encoding_argument(parser, "-e", "--output-encoding", "encoding of the vocabulary file")
  parser.add_argument("-u", "--unsorted", action="store_true",
                      help="do not sort vocabulary in output")
  add_deserialization_group(parser)
  add_mp_group(parser)
  return get_vocabulary_ns


def get_vocabulary_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  assert len(ns.dictionaries) > 0

  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  total_vocabulary = OrderedSet()
  for dictionary_path in ns.dictionaries:
    dictionary_instance = try_load_dict(
      dictionary_path, ns.deserialization_encoding, lp_options, mp_options, logger)
    if dictionary_instance is None:
      return False

    vocabulary = dictionary_instance.keys()
    total_vocabulary.update(vocabulary)

  sort = not ns.unsorted
  result = total_vocabulary
  if sort:
    result = sorted(result)
  content = "\n".join(result)

  try:
    ns.output.parent.mkdir(parents=True, exist_ok=True)
    cast(Path, ns.output).write_text(content, ns.output_encoding)
  except Exception as ex:
    logger.error("Vocabulary couldn't be saved!")
    logger.debug(ex)
    return False

  logger.info(f"Written vocabulary containing {len(result)} words to: \"{ns.output.absolute()}\".")
  return True



import logging
import os
from logging import DEBUG, Formatter, Handler, Logger, LogRecord, StreamHandler, getLogger
from logging.handlers import QueueHandler
from pathlib import Path
from queue import Queue
from typing import Dict, Generator, List, Tuple

from ordered_set import OrderedSet

# class StoreRecordsHandler(Handler):
# slower than other impl (maybe due to lock-things)
#   def __init__(self, level: int = DEBUG) -> None:
#     super().__init__(level)
#     self.__records = []

#   def handle(self, record) -> None:
#     self.__records.append(record)

#   def emit(self, record: LogRecord) -> None:
#     pass

#   @property
#   def records(self) -> List[LogRecord]:
#     return self.__records


class StoreRecordsHandler():
  def __init__(self) -> None:
    self.__records = []
    self.level = DEBUG

  def handle(self, record) -> None:
    self.__records.append(record)

  @property
  def records(self) -> List[LogRecord]:
    return self.__records


class ConsoleFormatter(logging.Formatter):
  """Logging colored formatter, adapted from https://stackoverflow.com/a/56944256/3638629"""

  purple = '\x1b[34m'
  blue = '\x1b[36m'
  # blue = '\x1b[38;5;39m'
  yellow = '\x1b[38;5;226m'
  red = '\x1b[1;49;31m'
  bold_red = '\x1b[1;49;31m'
  reset = '\x1b[0m'

  def __init__(self):
    super().__init__()
    self.datefmt = '%H:%M:%S'
    fmt = '(%(levelname)s) %(message)s'
    fmt_info = '%(message)s'

    self.fmts = {
        logging.NOTSET: self.purple + fmt + self.reset,
        logging.DEBUG: self.blue + fmt + self.reset,
        logging.INFO: fmt_info,
        logging.WARNING: self.yellow + fmt + self.reset,
        logging.ERROR: self.red + fmt + self.reset,
        logging.CRITICAL: self.bold_red + fmt + self.reset,
    }

  def format(self, record):
    log_fmt = self.fmts.get(record.levelno)
    formatter = logging.Formatter(log_fmt, self.datefmt)

    return formatter.format(record)


def add_console_out(logger: Logger):
  console = StreamHandler()
  logger.addHandler(console)
  set_console_formatter(console)


def init_and_return_loggers(name: str) -> Tuple[Logger, Logger]:
  logger = getLogger(name)
  flogger = get_file_logger()
  logger.parent = flogger
  logger.handlers.clear()
  assert len(logger.handlers) == 0
  add_console_out(logger)
  return flogger, logger


def init_file_stem_loggers(file_stems: OrderedSet[str]) -> Dict[str, Queue]:
  logging_queues = dict.fromkeys(file_stems)
  for k in file_stems:
    logger = getLogger(k)
    logger.propagate = False
    q = Queue(-1)
    logging_queues[k] = q
    handler = QueueHandler(q)
    logger.addHandler(handler)

  return logging_queues


def init_file_stem_logger_lists(file_stems: OrderedSet[str]) -> Dict[str, List[Tuple[int, str]]]:
  logging_queues = dict.fromkeys(file_stems)
  for k in file_stems:
    logging_queues[k] = []
  return logging_queues


def get_file_stem_loggers(file_stems: OrderedSet[str]) -> Generator[Logger, None, None]:
  for k in file_stems:
    logger = getLogger(k)
    yield logger


def write_file_stem_loggers_to_file_logger(queues: Dict[str, Queue]) -> None:
  flogger = get_file_logger()
  for k, q in queues.items():
    flogger.info(f"Log messages for file: {k}")
    entries = list(q.queue)
    for x in entries:
      flogger.handle(x)


def write_file_stem_logger_lists_to_file_logger(lists: Dict[str, List[Tuple[int, str]]]) -> None:
  flogger = get_file_logger()
  for k, l in lists.items():
    flogger.info(f"Log messages for file: {k}")
    for lvl, msg in l:
      flogger.log(lvl, msg)


def set_console_formatter(handler: Handler) -> None:
  logging_formatter = ConsoleFormatter()
  handler.setFormatter(logging_formatter)


def set_logfile_formatter(handler: Handler) -> None:
  fmt = '[%(asctime)s.%(msecs)03d] (%(levelname)s) %(message)s'
  datefmt = '%Y/%m/%d %H:%M:%S'
  logging_formatter = Formatter(fmt, datefmt)
  handler.setFormatter(logging_formatter)


def configure_root_logger() -> None:
  # productive = False
  # loglevel = logging.INFO if productive else logging.DEBUG
  main_logger = getLogger()
  main_logger.setLevel(logging.DEBUG)
  main_logger.manager.disable = logging.NOTSET
  if len(main_logger.handlers) > 0:
    console = main_logger.handlers[0]
  else:
    console = logging.StreamHandler()
    main_logger.addHandler(console)

  set_console_formatter(console)
  console.setLevel(logging.DEBUG)


def get_file_logger() -> Logger:
  logger = getLogger("file-logger")
  if logger.propagate:
    logger.propagate = False
  return logger


def try_init_file_logger(path: Path, level: int) -> bool:
  if path.is_dir():
    logger = getLogger(__name__)
    logger.error("Logging path is a directory!")
    return False
  flogger = get_file_logger()
  assert len(flogger.handlers) == 0
  try:
    path.parent.mkdir(parents=True, exist_ok=True)
    if path.is_file():
      os.remove(path)
    path.write_text("")
    fh = logging.FileHandler(path)
  except Exception as ex:
    logger = getLogger(__name__)
    logger.error("Logfile couldn't be created!")
    logger.exception(ex)
    return False

  set_logfile_formatter(fh)

  fh.setLevel(level)
  flogger.addHandler(fh)
  return True



from collections import OrderedDict

from ordered_set import OrderedSet

from pronunciation_dictionary_utils.pronunciations_map_symbols import (
  __init_pool, process_map_pronunciations_partial)


def test_component_change():
  dictionary = OrderedDict()
  pronunciations = OrderedDict()
  pronunciations[tuple(("t", "Ae", "s", "t"))] = 2
  pronunciations[tuple(("t", "Ae", "e", "s", "t"))] = 3
  pronunciations[tuple(("t", "s", "t"))] = 4  # keep untouched
  pronunciations[tuple(("t", "Ay", "s", "t"))] = 8  # same as first
  pronunciations[tuple(("t", "AX", "s", "t"))] = 5  # existing that will be extended

  dictionary["test"] = pronunciations

  __init_pool(dictionary)
  word, new_pronunciations = process_map_pronunciations_partial("test", OrderedSet(("e", "y")), "X")

  assert word == "test"
  assert new_pronunciations[tuple(("t", "AX", "s", "t"))] == 15
  assert new_pronunciations[tuple(("t", "AX", "X", "s", "t"))] == 3
  assert new_pronunciations[tuple(("t", "s", "t"))] == 4


def test_component_unchanged():
  dictionary = OrderedDict()
  pronunciations = OrderedDict()
  pronunciations[tuple(("t", "e", "s", "t"))] = 2
  dictionary["test"] = pronunciations

  __init_pool(dictionary)
  word, new_pronunciations = process_map_pronunciations_partial("test", OrderedSet(("X",)), "Y")

  assert word == "test"
  assert new_pronunciations is None



"""
The module maps symbols in pronunciations in *.dict file according to mappings from *.json file.

Example usage:
$ dict-cli map_symbols_in_pronunciations_json "my_dictionary.dict" "mappings.json" -pm

positional arguments:
  DICTIONARY                            dictionary file
  MAPPING                               mapping file

optional arguments:
  -h, --help                            show this help message and exit
  -pm, --partial-mapping                map symbols inside a symbol
"""

import json
from argparse import ArgumentParser, Namespace
from json.decoder import JSONDecodeError
from logging import Logger
from typing import Dict, Optional

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils.pronunciations_map_symbols_dict import map_symbols_dict
from pronunciation_dictionary_utils_cli.argparse_helper import (add_encoding_argument, add_io_group,
                                                                add_mp_group, parse_existing_file)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict

DEFAULT_EMPTY_WEIGHT = 1.0


def get_pronunciations_map_symbols_json_parser(parser: ArgumentParser) -> Namespace:
  """
  Configures an ArgumentParser object to parse command-line arguments.
  """
  parser.description = "Map symbols in pronunciations according to mappings in *.json file."
  parser.add_argument("dictionary", metavar='DICTIONARY',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("mapping", metavar='MAPPING',
                      type=parse_existing_file, help="mapping file")
  parser.add_argument("-pm", "--partial-mapping", action="store_true",
                      help="map symbols inside a symbol; useful when mapping the same vowel having different tones or stress within one operation")
  add_encoding_argument(parser, "-me", "--mapping-encoding", "encoding of mapping")
  add_io_group(parser)
  add_mp_group(parser)
  return map_symbols_in_pronunciations_ns


def try_load_mappings(flogger: Logger, mapping: str, encoding: str) -> Optional[Dict[str, str]]:
  with open(mapping, "r", encoding=encoding) as mapping_file:
    # warning: only last value saved for duplicate keys in file
    try:
      mappings = json.load(mapping_file)
    except (JSONDecodeError, TypeError):
      flogger.info("No or invalid content in mapping file.")
      return None
  if not isinstance(mappings, dict):
    flogger.info("No dictionary found in mapping file.")
    return None
  for key, value in mappings.items():
    if not isinstance(key, str) or not isinstance(value, str):
      flogger.info("Keys or values in mapping file are not of type string.")
      return None
  return mappings


def map_symbols_in_pronunciations_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  """
  Loads the dictionary and the mapping file and sends them for mapping, logs and saves the results.
  """
  lp_options = DeserializationOptions(
    ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)
  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options, logger)
  if dictionary_instance is None:
    return False
  logger.info(f"Loaded dictionary containing {len(dictionary_instance)} entries.")

  mappings = try_load_mappings(flogger, ns.mapping, ns.encoding)
  if mappings is None:
    return False
  logger.info(f"Loaded mapping containing {len(mappings)} entries.")

  changed_words_total = map_symbols_dict(
    dictionary_instance, mappings, ns.partial_mapping, mp_options, silent=False)

  if len(changed_words_total) == 0:
    logger.info("Didn't change anything.")
    return True

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options, logger)
  if not success:
    return False

  logger.info(f"Changed pronunciations of {len(changed_words_total)} word(s).")
  logger.info(f"Written dictionary to: \"{ns.dictionary.absolute()}\"")

  return True



from collections import OrderedDict
from functools import partial
from multiprocessing.pool import Pool
from typing import Optional, Tuple

from ordered_set import OrderedSet
from pronunciation_dictionary import MultiprocessingOptions, PronunciationDict, Pronunciations, Word
from tqdm import tqdm

from pronunciation_dictionary_utils.validation import (validate_dictionary, validate_mp_options,
                                                       validate_type)


def sort_pronunciations(dictionary: PronunciationDict, descending: bool, ignore_weight: bool, mp_options: MultiprocessingOptions, silent: bool = False) -> int:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_type(descending, bool):
    raise ValueError(f"Parameter 'descending': {msg}")
  if msg := validate_type(ignore_weight, bool):
    raise ValueError(f"Parameter 'ignore_weight': {msg}")
  if msg := validate_mp_options(mp_options):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  process_method = partial(
    process_sort_pronunciations,
    descending=descending,
    ignore_weight=ignore_weight,
  )

  with Pool(
    processes=mp_options.n_jobs,
    initializer=__init_pool_prepare_cache_mp,
    initargs=(dictionary,),
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    entries = OrderedSet(dictionary.keys())
    iterator = pool.imap(process_method, entries, mp_options.chunksize)
    pronunciations_to_i = list(tqdm(iterator, total=len(entries), unit="words", disable=silent))

  changed_counter = 0
  for word, new_pronunciations in pronunciations_to_i:
    pronunciations_unchanged = new_pronunciations is None
    if pronunciations_unchanged:
      continue

    dictionary[word] = new_pronunciations
    changed_counter += 1

  return changed_counter


process_lookup_dict: PronunciationDict = None


def __init_pool_prepare_cache_mp(lookup_dict: PronunciationDict) -> None:
  global process_lookup_dict
  process_lookup_dict = lookup_dict


def process_sort_pronunciations(word: Word, descending: bool, ignore_weight: bool) -> Tuple[Word, Optional[Pronunciations]]:
  global process_lookup_dict
  assert word in process_lookup_dict
  pronunciations = process_lookup_dict[word]
  new_pronunciations = sort_pronunciations_entry(pronunciations, descending, ignore_weight)
  changed_anything = pronunciations != new_pronunciations
  if changed_anything:
    return word, new_pronunciations
  return word, None


def sort_pronunciations_entry(pronunciations: Pronunciations, descending: bool, ignore_weight: bool) -> Pronunciations:
  pronunciations_sorted = sorted(pronunciations.items(
  ), key=lambda kv: kv[0] if ignore_weight else (-kv[1], kv[0]), reverse=descending)
  new_pronunciations = OrderedDict(pronunciations_sorted)
  return new_pronunciations



from argparse import ArgumentParser, Namespace
from logging import Logger
from pathlib import Path
from tempfile import gettempdir

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils import remove_symbols_from_words
from pronunciation_dictionary_utils_cli.argparse_helper import (ConvertToOrderedSetAction,
                                                                add_io_group, add_mp_group,
                                                                get_optional, parse_existing_file,
                                                                parse_path)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict


def get_words_remove_symbols_parser(parser: ArgumentParser):
  # TODO make it possible to include weights in output if they were not in input
  default_removed_out = Path(gettempdir()) / "removed-words.txt"
  parser.description = "Remove symbols from words. If all symbols of a word will be removed, the word will be taken out of the dictionary."
  parser.add_argument("dictionary", metavar='DICTIONARY',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("symbols", type=str, metavar='SYMBOL', nargs='+',
                      help="remove these symbols from the words", action=ConvertToOrderedSetAction)
  parser.add_argument("-m", "--mode", type=str, choices=["all", "start", "end", "both"], metavar="MODE",
                      help="mode to remove the symbols: all = on all locations; start = only from start; end = only from end; both = start + end", default="both")
  # parser.add_argument("--remove-empty", action="store_true",
  #                     help="if a pronunciation will be empty after removal, remove the corresponding word from the dictionary")
  parser.add_argument("-ro", "--removed-out", metavar="PATH", type=get_optional(parse_path),
                      help="write removed words to this file", default=default_removed_out)
  add_io_group(parser)
  add_mp_group(parser)
  return remove_symbols_from_words_ns


def remove_symbols_from_words_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  symbols_str = ''.join(ns.symbols)

  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options, logger)
  if dictionary_instance is None:
    return False

  removed_words_entirely, removed_words = remove_symbols_from_words(
    dictionary_instance, symbols_str, ns.mode, mp_options, silent=False)

  if len(removed_words) == 0:
    logger.info("Didn't change anything.")
    return True

  logger.info(f"Renamed {len(removed_words)} word(s).")

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options, logger)
  if not success:
    return False

  logger.info(f"Written dictionary to: \"{ns.dictionary.absolute()}\".")

  if len(removed_words_entirely) > 0:
    logger.warning(f"{len(removed_words_entirely)} words were removed entirely.")
    if ns.removed_out is not None:
      content = "\n".join(removed_words_entirely)
      ns.removed_out.parent.mkdir(parents=True, exist_ok=True)
      try:
        ns.removed_out.write_text(content, "UTF-8")
      except Exception as ex:
        logger.debug(ex)
        logger.error("Removed words output couldn't be created!")
        return False
      logger.info(f"Written removed words to: \"{ns.removed_out.absolute()}\".")
  else:
    logger.info("No words were removed.")
  return True



from collections import OrderedDict

import pytest
from pronunciation_dictionary import MultiprocessingOptions

from pronunciation_dictionary_utils.pronunciations_map_symbols_dict import map_symbols_dict


def test_with_changes() -> None:
  test_dictionary = OrderedDict([("test", OrderedDict([
                                  (("AO", "AO2", "AO3", "AA1", "EY2", "."), 1),
                                  (("A03", "NN", "HH"), 2)
                                  ]))])
  mappings = {
      "EY1": "ˈeɪ",
      "EY": "eɪ",
      "AO1": "ˈɔ",
      "EY0": "eɪ",
      "AO": "ɔ",
      "AO0": "ɔ",
      "AO2": "ˌɔ",
      "EY2": "ˌeɪ"
  }

  expected_result = OrderedDict([("test", OrderedDict([
                                  (("ɔ", "ˌɔ", "ɔ3", "AA1", "ˌeɪ", "."), 1),
                                  (("A03", "NN", "HH"), 2)]))])

  mp_options = MultiprocessingOptions(n_jobs=4, maxtasksperchild=100, chunksize=10)

  partial_mapping_flag = True
  changed_words = map_symbols_dict(
      test_dictionary, mappings, partial_mapping_flag, mp_options, silent=True)

  assert changed_words == {"test"}
  assert test_dictionary == expected_result


def test_with_whitespaces() -> None:
  test_dictionary = OrderedDict([("test", OrderedDict([(("EY2",), 1)]))])
  mappings = {"EY2": "ˌe ɪ"}

  mp_options = MultiprocessingOptions(n_jobs=4, maxtasksperchild=100, chunksize=10)

  partial_mapping_flag = True
  with pytest.raises(Exception) as expected_exception:
    map_symbols_dict(test_dictionary, mappings,
                     partial_mapping_flag, mp_options, silent=True)
  assert str(expected_exception.value) == "Whitespaces in mappings aren't supported with partial mapping."


def test_without_changes() -> None:
  test_dictionary = OrderedDict([("test", OrderedDict([
                                  (("AO", "AO2", "AO3", "AA1", "EY2", "."), 1),
                                  (("A03", "NN", "HH"), 2)
                                  ]))])
  mappings = {
      "V": "v",
      "F": "f",
      "D": "d",
      "MM": "m"
  }

  expected_result = OrderedDict([("test", OrderedDict([
                                  (("AO", "AO2", "AO3", "AA1", "EY2", "."), 1),
                                  (("A03", "NN", "HH"), 2)
                                  ]))])

  mp_options = MultiprocessingOptions(n_jobs=4, maxtasksperchild=100, chunksize=10)

  partial_mapping_flag = True
  changed_words = map_symbols_dict(test_dictionary, mappings,
                                   partial_mapping_flag, mp_options, silent=True)

  assert changed_words == set()
  assert test_dictionary == expected_result


def test_empty() -> None:
  test_dictionary = OrderedDict()
  mappings = dict()

  expected_result = OrderedDict()

  mp_options = MultiprocessingOptions(n_jobs=4, maxtasksperchild=100, chunksize=10)

  partial_mapping_flag = True
  changed_words = map_symbols_dict(test_dictionary, mappings,
                                   partial_mapping_flag, mp_options, silent=True)

  assert changed_words == set()
  assert test_dictionary == expected_result



from collections import OrderedDict

from pronunciation_dictionary import PronunciationDict, Pronunciations

from pronunciation_dictionary_utils.validation import validate_pronunciations


def merge_pronunciations(pronunciations1: Pronunciations, pronunciations2: Pronunciations) -> bool:
  if msg := validate_pronunciations(pronunciations1):
    raise ValueError(f"Parameter 'pronunciations1': {msg}")
  if msg := validate_pronunciations(pronunciations2):
    raise ValueError(f"Parameter 'pronunciations2': {msg}")

  if pronunciations1 == pronunciations2:
    return False

  changed_anything = False

  for pronunciation2, weight2 in pronunciations2.items():
    if pronunciation2 in pronunciations1:
      weight1 = pronunciations1[pronunciation2]
      new_weight = weight1 + weight2
      if weight1 != new_weight:
        pronunciations1[pronunciation2] = new_weight
        changed_anything = True
    else:
      pronunciations1[pronunciation2] = weight2
      changed_anything = True
  return changed_anything


# def merge_rel_weights(pronunciations1: Pronunciations, pronunciations2: Pronunciations, weights_ratio: float) -> bool:
#   if msg := validate_pronunciations(pronunciations1):
#     raise ValueError(f"Parameter 'pronunciations1': {msg}")
#   if msg := validate_pronunciations(pronunciations2):
#     raise ValueError(f"Parameter 'pronunciations2': {msg}")
#   if msg := validate_ratio(weights_ratio):
#     raise ValueError(f"Parameter 'weights_ratio': {msg}")

#   if pronunciations1 == pronunciations2:
#     return False

#   changed_anything = False
#   changed_anything |= convert_weights_to_probabilities(pronunciations1)
#   changed_anything |= convert_weights_to_probabilities(pronunciations2)

#   if weights_ratio != 1:
#     for pronunciation1, weight1 in pronunciations1.items():
#       if pronunciation1 not in pronunciations2:
#         new_weight = weight1 * weights_ratio
#         pronunciations1[pronunciation1] = new_weight
#         changed_anything = True

#   for pronunciation2, weight2 in pronunciations2.items():
#     new_weight2 = weight2 * (1 - weights_ratio)
#     if pronunciation2 in pronunciations1:
#       weight1 = pronunciations1[pronunciation2]
#       new_weight1 = weight1 * weights_ratio
#       new_weight = new_weight1 + new_weight2
#       if weight1 != new_weight:
#         pronunciations1[pronunciation2] = new_weight
#         changed_anything = True
#     else:
#       new_weight = new_weight2
#       pronunciations1[pronunciation2] = new_weight
#       changed_anything = True
#   return changed_anything


def convert_weights_to_probabilities_dict(dictionary: PronunciationDict) -> None:
  for pronunciations in dictionary.values():
    convert_weights_to_probabilities(pronunciations)


def convert_weights_to_probabilities(pronunciations: Pronunciations) -> bool:
  assert isinstance(pronunciations, OrderedDict)
  changed_anything = False
  sum_probs = sum(pronunciations.values())
  for pronunciation, prob in pronunciations.items():
    normed_prob = prob / sum_probs
    if prob != normed_prob:
      pronunciations[pronunciation] = normed_prob
      changed_anything = True
  return changed_anything


# def merge_equal_weight_sum(pronunciations1: Pronunciations, pronunciations2: Pronunciations, ratio: float) -> None:
#   assert sum(pronunciations1.values()) * ratio == sum(pronunciations2.values()) * (1 - ratio)
#   for pronunciation2, weight2 in pronunciations2.items():
#     if pronunciation2 in pronunciations1:
#       pronunciations1[pronunciation2] += weight2
#     else:
#       pronunciations1[pronunciations2] = weight2



import re
from collections import OrderedDict

from pronunciation_dictionary_utils.pronunciations_replace_pronunciation import \
  replace_symbols_from_pronunciations_entry


def test_component():
  pronunciations = OrderedDict((
    (("ˌə", "r", "r"), 6),
    (("ˌə", "ˈɔ", "r"), 7),
    (("ˌə", "ʊ", "r"), 8),
    (("ˌə", "ʊr"), 9),
  ))

  pattern = re.compile(r"( |ˈ|ˌ)(ə|ʌ|ɔ|ɪ|ɛ|ʊ) r")
  result = replace_symbols_from_pronunciations_entry(pronunciations, pattern, r"\1\2r")

  assert result == OrderedDict((
    (('ˌər', 'r'), 6),
    (('ˌə', 'ˈɔr'), 7),
    (('ˌə', 'ʊr'), 17),
  ))


def test_empty():
  pronunciations = OrderedDict((
    (("ˌə", "r", "r"), 6),
    (("ˌə", "ʊr"), 9),
  ))

  pattern = re.compile(r".*")
  result = replace_symbols_from_pronunciations_entry(pronunciations, pattern, r"")

  assert result == OrderedDict()


def test_whitespace():
  pronunciations = OrderedDict((
    (("ˌə", "r", "r"), 6),
    (("ˌə", "ʊr"), 9),
  ))

  pattern = re.compile(r".*")
  result = replace_symbols_from_pronunciations_entry(pronunciations, pattern, r" ")

  assert result == OrderedDict()



from argparse import ArgumentParser, Namespace
from logging import Logger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions)

from pronunciation_dictionary_utils import change_word_casing
from pronunciation_dictionary_utils_cli.argparse_helper import (add_io_group, add_mp_group,
                                                                parse_existing_file)
from pronunciation_dictionary_utils_cli.io import try_load_dict, try_save_dict


def get_words_casing_adjustment_parser(parser: ArgumentParser):
  parser.description = "Adjust casing of words in dictionary."
  parser.add_argument("dictionary", metavar='DICTIONARY',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("mode", metavar="MODE", type=str, choices=["lower", "upper"],
                      help="mode to change the casing")
  add_io_group(parser)
  add_mp_group(parser)
  return change_casing_ns


def change_casing_ns(ns: Namespace, logger: Logger, flogger: Logger) -> bool:
  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options, logger)
  if dictionary_instance is None:
    return False

  removed_words, created_words = change_word_casing(dictionary_instance, ns.mode, mp_options, silent=False)

  if len(removed_words) == 0:
    logger.info("Didn't change anything.")
    return True

  logger.info(f"Replaced {len(removed_words)} with {len(created_words)} word spelling(s).")

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options, logger)
  if not success:
    return False

  logger.info(f"Written dictionary to: \"{ns.dictionary.absolute()}\".")
  return True



from functools import partial
from logging import getLogger
from typing import Dict, Set

from ordered_set import OrderedSet
from pronunciation_dictionary import MultiprocessingOptions, get_phoneme_set
from pronunciation_dictionary.types import PronunciationDict
from tqdm import tqdm

from pronunciation_dictionary_utils import map_symbols


def get_mappable_symbols(sounds_in_dictionary: Set[str], sounds_in_mappings: Set[str]) -> Set[str]:
  """
  Returns a list of unique symbols that occur both in the passed dictionary and the mappings. 
  """
  mappable_symbols = sounds_in_dictionary & sounds_in_mappings
  return mappable_symbols


def get_unmappable_symbols(sounds_in_dictionary: Set[str], sounds_in_mappings: Set[str]) -> Set[str]:
  """
  Returns a list of unique symbols that occur in the dictionary but not in the mappings.
  """
  unmappable_symbols = sounds_in_dictionary - sounds_in_mappings
  return unmappable_symbols


def apply_mapping_partial(dictionary: PronunciationDict, mappings: Dict[str, str], mappable_symbol: str,
                          mp_options: 'MultiprocessingOptions') -> Set[str]:
  """
  Applies partial mapping for a single mappable symbol. It returns a set of words whose pronunciations 
  have been changed. It enables mapping the symbol occuring within a longer symbol, e.g., passing the 
  symbol "AO1" and a mapping "AO": "ɔ" would result in "ɔ1".
  """
  from_symbol = OrderedSet((mappable_symbol,))
  to_phonemes = mappings[mappable_symbol]

  if " " in to_phonemes:
    raise Exception("Whitespaces in mappings aren't supported with partial mapping.")

  changed_words = map_symbols(dictionary, from_symbol, to_phonemes,
                              True, mp_options, silent=True)

  return changed_words


def apply_mapping_full(dictionary: PronunciationDict, mappings: Dict[str, str], mappable_symbol: str,
                       mp_options: 'MultiprocessingOptions') -> Set[str]:
  """
  Applies mapping for a single mappable symbol. It returns a set of words whose pronunciations have 
  been changed. It does not map the symbol occuring within a longer symbol e.g., passing the 
  symbol "AO1" and a mapping "AO": "ɔ" would result in "AO1".
  """
  from_symbol = OrderedSet((mappable_symbol,))
  to_phonemes = mappings[mappable_symbol]

  # method without partial flag allows whitespaces within mappings
  to_phonemes = to_phonemes.split(" ")
  to_phonemes = [p for p in to_phonemes if len(p) > 0]

  changed_words = map_symbols(dictionary, from_symbol, to_phonemes,
                              False, mp_options, silent=True)

  return changed_words


def map_symbols_dict(dictionary: PronunciationDict, mappings: Dict[str, str],
                     partial_mapping: bool, mp_options: MultiprocessingOptions, silent: bool = False) -> Set[str]:
  """
  Identifies applicable mappings and triggers mapping them according to the specified method.
  It returns a set of words whose pronunciations have been changed. 
  """
  logger = getLogger(__name__)
  unique_sounds_in_dictionary = get_phoneme_set(dictionary)
  unique_sounds_in_mappings = set(mappings.keys())

  mappable_symbols = get_mappable_symbols(unique_sounds_in_dictionary, unique_sounds_in_mappings)
  unmappable_symbols = get_unmappable_symbols(
    unique_sounds_in_dictionary, unique_sounds_in_mappings)

  logger.info(f"Found {len(mappable_symbols)} applicable phoneme mappings.")
  logger.info(f"Mapped phonemes in dictionary: {' '.join(sorted(mappable_symbols))}")
  logger.info(f"Unmapped phonemes in dictionary: {' '.join(sorted(unmappable_symbols))}")

  changed_words_total = set()
  if mappable_symbols:
    mapping_method = partial(
      apply_mapping_partial if partial_mapping else apply_mapping_full,
      dictionary=dictionary,
      mappings=mappings,
      mp_options=mp_options
    )

    sorted_mappable_symbols = OrderedSet(sorted(mappable_symbols, key=len, reverse=True))

    for mappable_symbol in tqdm(sorted_mappable_symbols, total=len(sorted_mappable_symbols), desc="Mapping", disable=silent):
      changed_words = mapping_method(mappable_symbol=mappable_symbol)
      changed_words_total |= changed_words

  return changed_words_total



from collections import OrderedDict
from functools import partial
from multiprocessing.pool import Pool
from typing import Generator, Iterable, Literal, Optional, Set, Tuple

from ordered_set import OrderedSet
from pronunciation_dictionary import (MultiprocessingOptions, PronunciationDict, Pronunciations,
                                      Symbol, Word)
from tqdm import tqdm

from pronunciation_dictionary_utils.validation import (validate_dictionary, validate_mp_options,
                                                       validate_type)

DEFAULT_EMPTY_WEIGHT = 1


def __validate_mode(mode: str) -> Optional[str]:
  if mode not in ["all", "start", "end", "both"]:
    return "Value needs to be 'all', 'start', 'end' or 'both'!"
  return None


def remove_symbols_from_pronunciations(dictionary: PronunciationDict, symbols: OrderedSet[Symbol], mode: str, keep_empty: bool, empty_symbol: Optional[Symbol], mp_options: MultiprocessingOptions, silent: bool = False) -> Tuple[OrderedSet[Word], int]:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_type(symbols, OrderedSet):
    raise ValueError(f"Parameter 'symbols': {msg}")
  if msg := __validate_mode(mode):
    raise ValueError(f"Parameter 'mode': {msg}")
  if msg := validate_type(keep_empty, bool):
    raise ValueError(f"Parameter 'keep_empty': {msg}")
  if empty_symbol is not None and (msg := validate_type(empty_symbol, str)):
    raise ValueError(f"Parameter 'empty_symbol': {msg}")
  if msg := validate_mp_options(mp_options):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  if len(symbols) == 0:
    return OrderedSet(), 0

  process_method = partial(
    process_get_pronunciation,
    symbols=symbols,
    mode=mode,
  )

  with Pool(
    processes=mp_options.n_jobs,
    initializer=__init_pool_prepare_cache_mp,
    initargs=(dictionary,),
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    entries = OrderedSet(dictionary.keys())
    iterator = pool.imap(process_method, entries, mp_options.chunksize)
    pronunciations_to_i = list(tqdm(iterator, total=len(entries), unit="words", disable=silent))

  changed_counter = 0
  removed_words = OrderedSet()
  for word, new_pronunciations in pronunciations_to_i:
    pronunciations_unchanged = new_pronunciations is None
    if pronunciations_unchanged:
      continue

    if len(new_pronunciations) == 0:
      if keep_empty:
        assert empty_symbol is not None
        dictionary[word] = OrderedDict((
          ((empty_symbol,), DEFAULT_EMPTY_WEIGHT),
        ))
      else:
        removed_words.add(word)
        dictionary.pop(word)
    else:
      dictionary[word] = new_pronunciations
    changed_counter += 1

  return removed_words, changed_counter


process_lookup_dict: PronunciationDict = None


def __init_pool_prepare_cache_mp(lookup_dict: PronunciationDict) -> None:
  global process_lookup_dict
  process_lookup_dict = lookup_dict


def process_get_pronunciation(word: Word, symbols: Set[Symbol], mode: str) -> Tuple[Word, Optional[Pronunciations]]:
  global process_lookup_dict
  assert word in process_lookup_dict
  pronunciations = process_lookup_dict[word]
  new_pronunciations = remove_symbols_from_pronunciations_entry(pronunciations, symbols, mode)
  if new_pronunciations == pronunciations:
    return word, None
  return word, new_pronunciations


def remove_symbols_start(symbols: Iterable[Symbol], remove: Set[Symbol]) -> Generator[Symbol, None, None]:
  start = True
  for s in symbols:
    if start and s in remove:
      continue
    start = False
    yield s


def remove_symbols_end(symbols: Iterable[Symbol], remove: Set[Symbol]) -> Generator[Symbol, None, None]:
  symbols_rev = reversed(tuple(symbols))
  symbols_rev_rem = tuple(remove_symbols_start(symbols_rev, remove))
  result = reversed(symbols_rev_rem)
  yield from result


def remove_symbols_both(symbols: Iterable[Symbol], remove: Set[Symbol]) -> Generator[Symbol, None, None]:
  step1 = remove_symbols_start(symbols, remove)
  yield from remove_symbols_end(step1, remove)


def remove_symbols_all(symbols: Iterable[Symbol], remove: Set[Symbol]) -> Generator[Symbol, None, None]:
  res = (
    symbol
    for symbol in symbols
    if symbol not in remove
  )
  yield from res


def remove_symbols_mode(symbols: Iterable[Symbol], remove: Set[Symbol], mode: Literal["all", "start", "end", "both"]) -> Generator[Symbol, None, None]:
  if mode == "all":
    yield from remove_symbols_all(symbols, remove)
  elif mode == "start":
    yield from remove_symbols_start(symbols, remove)
  elif mode == "end":
    yield from remove_symbols_end(symbols, remove)
  elif mode == "both":
    yield from remove_symbols_both(symbols, remove)
  else:
    assert False


def remove_symbols_from_pronunciations_entry(pronunciations: Pronunciations, symbols: Set[Symbol], mode: str) -> Pronunciations:
  new_pronunciations = OrderedDict()
  changed_anything = False
  for pronunciation, weight in pronunciations.items():
    new_pronunciation = tuple(remove_symbols_mode(pronunciation, symbols, mode))

    if new_pronunciation != new_pronunciations:
      changed_anything = True

    if len(new_pronunciation) > 0:
      if new_pronunciation in new_pronunciations:
        new_pronunciations[new_pronunciation] += weight
      else:
        new_pronunciations[new_pronunciation] = weight
  if changed_anything:
    return new_pronunciations
  return pronunciations
